@@ -1,5 +1,4 @@
-#
-#  Copyright 2014 MongoDB, Inc.
+#  Copyright 2014 - present MongoDB, Inc.
 #
 #  Licensed under the Apache License, Version 2.0 (the "License");
 #  you may not use this file except in compliance with the License.
@@ -12,19 +11,18 @@
 #  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 #  See the License for the specific language governing permissions and
 #  limitations under the License.
-#
 
+use strict;
+use warnings;
 package MongoDB::Role::_ReadOp;
 
 # MongoDB role for read ops that provides read preference
 
 use version;
-our $VERSION = 'v1.2.2';
+our $VERSION = 'v2.2.2';
 
 use Moo::Role;
 
-use MongoDB::ReadPreference;
-use MongoDB::_Constants;
 use MongoDB::_Types qw(
     ReadPreference
     ReadConcern
@@ -32,20 +30,21 @@ use MongoDB::_Types qw(
 use Types::Standard qw(
     Maybe
 );
-use namespace::clean;
 
-with 'MongoDB::Role::_DatabaseOp';
+use namespace::clean;
 
 # PERL-573 Would like to refactor to remove Maybe types for
 # read_preference and read_concern
 has read_preference => (
-    is  => 'ro',
-    isa => Maybe [ReadPreference],
+    is  => 'rw', # rw for Op::_Query which can be modified by Cursor
+    required => 1,
+    isa => ReadPreference,
 );
 
 has read_concern => (
     is  => 'ro',
-    isa => Maybe [ReadConcern],
+    required => 1,
+    isa => ReadConcern,
 );
 
 1;
